/*
 * title:       market.c
 *
 * author:      Craig E. Ward
 *
 * envionment:  UNIX, ANSI C (gcc 3.1)
 *
 * description:
 *              CMSI 587 Operating Systems, Spring 2003
 *              Term Project
 *              Instructor: Tri Nguyen
 *              Due: April 30
 *
 *
 */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>

#include "mytypes.h"            /* generic application types */
#include "protocol.h"           /* messaging protocol */
#include "stocks.h"             /* stock structures */
#include "transport.h"          /* the message transport module */
#include "util.h"               /* utility routines */

#include <pthread.h>

/* Number of trades between displays of the market */
#define DEFAULT_DISPLAY_INTERVAL	4

/* static function prototypes */
static void show_market(void);
static void init_market(void);
static void run_server(void);
static void process_request(request_data_t *);

/*  external data  */

/* initialized data */
static int stocks_in_market = SYMBOL_COUNT;
static int display_interval = DEFAULT_DISPLAY_INTERVAL;
static unsigned int request_counter = 0;

/* uninitialized data */
static char *progname;
static struct listing *market;
static pthread_mutex_t socket_mutex = PTHREAD_MUTEX_INITIALIZER;

int main (int argc, char *argv[])
{
  progname = strrchr(argv[0], '/');
  if (progname == NULL) { progname = argv[0]; } else { progname++; }

  if (argc > 1) {
    display_interval = atoi(argv[1]);
    if (display_interval <= 0) display_interval = DEFAULT_DISPLAY_INTERVAL;
  }

  if (start_server() == TRUE) {
    run_server();
  } else {
    (void)stop_server();
    exit(1);
  }

  (void)stop_server();
  exit(0);
} /* main */

static void init_market()
{
  int i = 0;

  seed_random_choice((unsigned long)time(0));

  market = (struct listing *)malloc(sizeof(struct listing));
  memset(market, '\0', sizeof(struct listing));
  struct listing *current = market;

  for (i = 0; i < NUMBER_SYMBOLS; i++) {
    current->astock = (struct stock *)malloc(sizeof(struct stock));
    memset(current->astock, '\0', sizeof(struct stock));
    strncpy(current->astock->symbol, symbols[i], SYMBOL_LENGTH);
    current->astock->unitvalue = 9929; /* $99.29 in cents! */
    current->astock->code = i;
    current->available_shares = 1100;
    current->next = malloc(sizeof(struct listing));
    current = current->next;
    current->next = 0;
    /* initialize this mutex with default values */
    pthread_mutex_init(&(current->stock_mutex), NULL);
  }
} /* init_market */

/*
 * Traverse the linked list representing the market and display the current
 * values for non-delisted issues.
 */
static void show_market()
{
  int i = 0;
  struct listing *listings = market;
  time_t now = time(0);

  printf("============================================================\n");
  printf("Market Listings as of %s\n", ctime(&now));
  for (i = 0; i < NUMBER_SYMBOLS; i++) {
    if (listings->available_shares != DELISTED_FLAG) {
      printf("Stock: %s Available: %5d Unit Value: $%8.2f\n",
             listings->astock->symbol,
             listings->available_shares,
             (listings->astock->unitvalue/100));
    }
    listings=listings->next;
  }
  printf("============================================================\n");
  fflush(stdout);
} /* show_market */

static void run_server()
{
  pthread_t *worker_thread_p;
  request_data_t *request_p;
  int status;

  init_market();
  printf("%s: server started; displaying market every %d trades\n",
	 progname, display_interval);
  show_market(); 

  while (stocks_in_market > 0) {
    request_counter++;		/* count trade for display purposes */
    
    request_p = get_request();
    if (request_p == NULL) {
      log("market.run_server: abort!");
      break;
    }
    request_p->request_number = request_counter;
    worker_thread_p = (pthread_t *)malloc(sizeof(pthread_t));
    status = pthread_create(worker_thread_p,
			    NULL,
			    (void *)process_request,
			    (void *)request_p);
    if (status != 0) {
      log("%s.run_server: pthread_create failed: %s",
	  __FILE__, strerror(status));
      break;
    }
    /* 
     * Mark this thread for deletion so the library will cleanup after it
     * when it exits without telling us and free what we allocated here.
     */
    pthread_detach(*worker_thread_p);
    free(worker_thread_p);	/* return storage */

    if ((request_counter % display_interval) == 0) {
      show_market();
    }
  } /* while (stocks_in_market > 0) */

  if (stocks_in_market == 0) {
    log("market.run_server: stocks_in_market = %d; request_counter = %d",
	stocks_in_market, request_counter);
    printf("\n%s: no more stocks on the exchange after %d requests,"
	   " shutting down.\n\n",
           progname, request_counter);
  } else {
    log("%s: market halting due to error: stocks_in_market = %d;"
	" request_counter = %d", 
	progname, stocks_in_market, request_counter);
    printf("%s: market halting due to error: stocks_in_market = %d;"
	   " request_counter = %d", 
	   progname, stocks_in_market, request_counter);
  }
  return;
}

/*
 * update the unit value, share price, of a stock
 */
static double update_share_price(double old_price)
{
  log("market.update_share_price: old_price=%.0f", old_price);
  int choice = random_choice(2);
  double new_price;
  switch (choice)
    {
    case 0:
      new_price = old_price - (old_price * .25);
      break;
    default:
      new_price = old_price + (old_price * .25);
      break;
    }
  log("market.update_share_price: new_price=%.0f", new_price);
  return new_price;
}


/*
 * process buying of a stock
 */
static void process_stock_buy(int symbol_code, char *msgbuf)
{
  log("market.process_stock_buy: symbol_code=%d (%s)", 
      symbol_code, symbols[symbol_code]);
  struct listing *alisting;

  alisting = get_stock_listing(market, symbol_code);

  if (alisting != NULL) {

    pthread_mutex_lock(&(alisting->stock_mutex));
    if (alisting->available_shares >= 100) {
      alisting->astock->unitvalue = 
        update_share_price(alisting->astock->unitvalue);
      if (alisting->astock->unitvalue < MIN_UNIT_VALUE) {
        /* mark as delisted and send bad news */
        alisting->available_shares = DELISTED_FLAG;
        sprintf(msgbuf,"DELIST:%s", symbols[symbol_code]);
        stocks_in_market--;
      } else {
        sprintf(msgbuf,"BUY:YES:%.0f", alisting->astock->unitvalue);
        alisting->available_shares = alisting->available_shares - 100;
      }
    } else if (alisting->available_shares < 0) {
      /* negitive is delisted indicator */
      sprintf(msgbuf,"DELIST:%s", symbols[symbol_code]);
    } else {
      sprintf(msgbuf,"BUY:NO");      
    }
    pthread_mutex_unlock(&(alisting->stock_mutex));

  } else {
    log("ERROR in market.process_stock_buy: can find listing for symbo"
        "l_code=%d, symbol=%s",
        symbol_code, symbols[symbol_code]);
    sprintf(msgbuf,"ERROR in market.process_stock_buy: can find listing "
            "for symbol_code=%d, symbol=%s",
            symbol_code, symbols[symbol_code]);
  }
  return;
}

/*
 * process selling of a stock
 */
static void process_stock_sale(int symbol_code, char *msgbuf)
{
  log("market.process_stock_sale: symbol_code=%d (%s)", 
      symbol_code, symbols[symbol_code]);
  struct listing *alisting;

  alisting = get_stock_listing(market, symbol_code);

  if (alisting != NULL) {

    pthread_mutex_lock(&(alisting->stock_mutex));
    switch (alisting->available_shares)
      {
      case -1:                  /* already delisted! */
        sprintf(msgbuf,"DELIST:%s", symbols[symbol_code]);
        break;
      default:
        alisting->astock->unitvalue = 
          update_share_price(alisting->astock->unitvalue);
        if (alisting->astock->unitvalue < MIN_UNIT_VALUE) {
          /* mark as delisted and send bad news */
          alisting->available_shares = -1;
          sprintf(msgbuf,"DELIST:%s", symbols[symbol_code]);
          stocks_in_market--;
        } else {
          sprintf(msgbuf,"SELL:YES");
          alisting->available_shares = alisting->available_shares + 100;
        }
      } /* switch */

    pthread_mutex_unlock(&(alisting->stock_mutex));

  } else {
    log("ERROR in market.process_stock_sale: can find listing for symbo"
        "l_code=%d, symbol=%s",
        symbol_code, symbols[symbol_code]);
    sprintf(msgbuf,"ERROR in market.process_stock_sale: can find listing "
            "for symbol_code=%d, symbol=%s",
            symbol_code, symbols[symbol_code]);
  }
  return;
}

/*
 * process check of a stock price
 */
static void process_price_check(int symbol_code, char *msgbuf)
{
  log("market.process_price_check: symbol_code=%d (%s)", 
      symbol_code, symbols[symbol_code]);
  struct listing *alisting;

  alisting = get_stock_listing(market, symbol_code);

  if (alisting != NULL) {

    pthread_mutex_lock(&(alisting->stock_mutex));
    if (alisting->available_shares < 0) {
      /* mark as delisted and send bad news */
      sprintf(msgbuf,"DELIST:%s", symbols[symbol_code]);
    }else {
      sprintf(msgbuf,"CHECK:YES:%.0f", 
              alisting->astock->unitvalue); /* success! */
    }

    pthread_mutex_unlock(&(alisting->stock_mutex));
  } else {
    log("ERROR in market.process_price_check: can find listing for symbo"
        "l_code=%d, symbol=%s",
        symbol_code, symbols[symbol_code]);
    sprintf(msgbuf,"ERROR in market.process_price_check: can find listing "
            "for symbol_code=%d, symbol=%s",
            symbol_code, symbols[symbol_code]);
  }
  return;
}

/*
 * Extract the parts of a request message and dispatch to appropriate handler.
 */
static void process_request(request_data_t *request_p)
{
  log("market.process_request: enter: request %d = %s", 
      request_p->request_number,request_p->request);
  bool_t status = TRUE;
  char *request_part;
  int request_code = 0;
  char *symbol_part;
  int symbol_code = 0;
  int request_number = request_p->request_number;

  /* parse the request */
  request_part = get_request_part(request_p->request);
  request_code = find_request_by_name(request_part);
  symbol_part = get_request_symbol_part(request_p->request);
  symbol_code = find_stock_by_name(symbol_part);

  /* dispatch to handlers */
  switch (request_code)
    {
    case BUY:
      process_stock_buy(symbol_code, request_p->result);
      break;
    case SELL:
      process_stock_sale(symbol_code, request_p->result);
      break;
    case CHECK:
      process_price_check(symbol_code, request_p->result);
      break;
    default:
      log("%s: unexpected request code: %d",
          progname, request_code);
      status = FALSE;
      break;
    }

  if (status == TRUE) {
#if 1
    pthread_mutex_lock(&socket_mutex);
    status = send_result(request_p);
    pthread_mutex_unlock(&socket_mutex);
#else
    status = send_result(request_p);
#endif
  }
  log("market.process_request: thread for request %d exit"
      " (stocks_in_market=%d)", request_number, stocks_in_market);

  return;			/* implicit pthread_exit()  */
}
/* */
