/* 
 * title:       Wrappers for semaphore system calls
 *
 * author:      Craig E. Ward
 *
 * environment: Mac OS X (Darwin/GCC 3.1)
 *
 * description:
 *              CMSI 587 Operating Systems, Spring 2003
 *              Assignment 4 -- Interprocess Communication &
 *              Synchronication-Coordination
 *              Instructor: Tri Nguyen
 *              Due: March 19
 *
 * The Darwin implementation of semaphores is slightly different from other
 * UNIX-class operating systems, such as Linux, Solaris, etc. While these
 * functions can be used on other systems, this supporting code can not.
 *
 * The main difference seems to be in naming. The sem_open() call here is
 * called sem_init() on some other systems.
 */
#include <errno.h>
#include <string.h>

#if USESVR4
#define __USE_SVID 	1
#define __SVID_SOURCE	1
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif

#include "util.h"
#include "semaphores.h"

#if USESVR4
static int sset_identifier = 0;	/* semaphore set identifier */
static key_t key = (key_t)561776;

/* The union for semctl may or may not be defined for us.  This code, defined
   in Linux's semctl() manpage, is the proper way to attain it if necessary. */

#if defined(__GNU_LIBRARY__) && !defined(_SEM_SEMUN_UNDEFINED)
/* union semun is defined by including <sys/sem.h> */
#else
/* according to X/OPEN we have to define it ourselves */
union semun {
  int val;                    /* value for SETVAL */
  struct semid_ds *buf;       /* buffer for IPC_STAT, IPC_SET */
  unsigned short int *array;  /* array for GETALL, SETALL */
  struct seminfo *__buf;      /* buffer for IPC_INFO */
};
#endif
#endif

/*
 * Create a semaphore
 */
sem_t *semaphore_util_create(const char *name)
{
#if USESVR4
  log(__FILE__ ".semaphore_create: SVR4 calls ignoring \"%s\"", name);
  union semun sunion;

  sset_identifier = semget(IPC_PRIVATE, 1, 00644);

  if (sset_identifier == -1) {
    log(__FILE__ ".semaphore_create: semget failed: \"%s\"", strerror(errno));
  } else {
    /* Initialize its resource count to 1. */
    sunion.val = 1;
    if ((semctl(sset_identifier, 0, SETVAL, sunion)) == -1) {
      log(__FILE__ ".semaphore_create (SVR4) semctl failed: %s",
	  strerror(errno));
      sset_identifier = -1;
    }
  }
  return &sset_identifier;
#else
  mode_t mode = 0644;
  sem_t *semaphore;
  log(__FILE__ ".semaphore_create: \"%s\"", name);
  semaphore = sem_open(name, O_CREAT, mode, 1);
  log(__FILE__ ".semaphore_create: after call to sem_open");
  if (semaphore == (sem_t *)SEM_FAILED) {
    log(__FILE__ ".semaphore_create: create failed: %s", strerror(errno));
  }
  log(__FILE__ ".semaphore_create: return");
  return semaphore;
#endif
}

/*
 * Mark the semaphore as available
 */
int semaphore_release(sem_t *semaphore)
{
  int status = 0;

#if USESVR4
  log(__FILE__ ".semaphore_release (SVR4): releasing semaphore id %d",
      *semaphore);
  struct sembuf ops = {0, 1, 0};
  status = semop(*semaphore, &ops, 1);
#else
  log(__FILE__".semaphore_release: enter");
  status = sem_post(semaphore);
#endif

  if (status != 0) {
    log(__FILE__ ".semaphore_release: release failed: %s", strerror(errno));
  }
  log(__FILE__".semaphore_release: exit");
  return status;
}

/*
 * Acquire the semaphone, blocking if necessary
 */
int semaphore_acquire(sem_t *semaphore)
{
  int status = 0;
#if USESVR4
  log(__FILE__ ".semaphore_acquire (SVR4): releasing semaphore id %d",
      *semaphore);
  struct sembuf ops = {0, -1, 0};
  status = semop(*semaphore, &ops, 1);
#else
  log(__FILE__".semaphore_acquire: enter");
  status = sem_wait(semaphore);
#endif
 if (status != 0) {
    log(__FILE__ ".semaphore_acquire: acquire failed: %s", strerror(errno));
  }
  log(__FILE__".semaphore_acquire: exit");
  return status;
}

/*
 * Remove the semaphore from the system
 */
int semaphore_close(sem_t *semaphore)
{
  log(__FILE__".semaphore_close: enter");
  int status = 0;
#if USESVR4
  status = semctl(*semaphore, 0, IPC_RMID, 0);
#else
  status = sem_close(semaphore);
#endif
  if (status != 0) {
    log(__FILE__ ".semaphore_close: acquire failed: %s", strerror(errno));
  }
  log(__FILE__".semaphore_close: return: status=%d",status);
  return status;
}
