/*
 * title:       trader.c
 *
 * author:      Craig E. Ward
 *
 * envionment:  UNIX, ANSI C (gcc 3.1)
 *
 * description:
 *              CMSI 587 Operating Systems, Spring 2003
 *              Term Project -- Market Simulation
 *              Instructor: Tri Nguyen
 *              Due April 30
 *
 */
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>

#include "mytypes.h"            /* generic application types */
#include "protocol.h"           /* messaging protocol */
#include "stocks.h"             /* stock structures */
#include "transport.h"          /* message transport */
#include "util.h"               /* utility routines */
#include "semaphores.h"         /* my wrappers */

#define DEFAULT_NUMBER_TRADERS  10
#define TRADER_SLEEP_SECONDS    3

/* function prototypes */
static void startup_trader(int, const char *);
static void begin_trading(void);
static void update_portfolio(int , int, char *);
static int find_stock_code(int);
static bool_t do_sell(void);
static bool_t do_buy(void);
static bool_t do_check(void);
static void do_display(void);
static void process_buy(int, char *);
static void process_sell(int, char *);
static void process_check(int, char *);
static void usage(void);

/*  external data  */
extern char *optarg;
extern int optind;
extern int optopt;
extern int opterr;
extern int optreset;

/* initialized data */
static const char SEM_NAME[] = "DISPLAY";
static int trader_number = 1;
static int number_traders = DEFAULT_NUMBER_TRADERS;
static int stock_count = 0;
static struct portfolio_entry *portfolio = NULL;


/* uninitialized data */
char *progname;
static char server_dotted_decimal_address[MAX_DOTTED_DECIMAL + 1];
static sem_t *semaphore;
static char result[MAX_REQUEST_LENGTH +
		   MAX_RESULT_LENGTH  +
		   MAX_PRICE_STRING   + 1];

int main (int argc, char *argv[])
{
  pid_t pid;
  int i, ch;

  progname = strrchr(argv[0], '/');
  if (progname == NULL) { progname = argv[0]; } else { progname++; }

  strncpy(server_dotted_decimal_address, DEFAULT_SERVER_ADDR,
	 sizeof server_dotted_decimal_address);

  while ((ch = getopt(argc, argv, "h:t:")) != -1) {
    switch (ch) {
    case 'h':
      strncpy(server_dotted_decimal_address, optarg,
	      sizeof server_dotted_decimal_address);
      log("trader.main: set server address to %s",
	  server_dotted_decimal_address);
      break;

    case 't':
      number_traders = atoi(optarg);
      log("trader.main: set number of traders to %d",
	  number_traders);
      if (number_traders <= 0) 
	number_traders = DEFAULT_NUMBER_TRADERS;
      break;

    case '?':
    default:
      usage();
      exit(EXIT_FAILURE);
    }
  }

  semaphore = semaphore_util_create(SEM_NAME);

  /* fork however many trader child processes desired. default is
     DEFAULT_NUMBER_TRADERS
     ------------------------------------------------------------ */
  for (i = 1; i < number_traders; i++) {
    pid = fork();
    switch (pid) 
      {
      case -1:
        log("%s: fork error: %s", strerror(errno));
        exit(2);
        /* NOT REACHED */
        break;
      case 0:
        startup_trader(i, server_dotted_decimal_address);
        break;
      default:
        log("%s: created child trader pid = %d", progname, pid);
        sleep(2);               /* get different seeds */
        break;
      }
  }
  log("%s: parent starting trading",progname);
  startup_trader(0,  server_dotted_decimal_address);
  /* NOTREACHED */
  exit(EXIT_SUCCESS);
} /* main */

/*
 * Startup the trader. This function does not return.
 */
static void startup_trader(int number, const char *server_address)
{
  log(__FILE__ ".startup_trader number %d for server %s",
      number,  server_dotted_decimal_address);
  printf("%s: starting up Trader Number %d\n",progname, number);
  /* initalize random number sequence */
  seed_random_choice((unsigned long)time(0));
  trader_number = number;
  if (init_trader(server_address) == TRUE) {
    begin_trading();
    log("%s: begin_trading has returned", progname);
  } else {
    log("%s: can not initialize transport", progname);
    exit(1);
  }
  if (stop_trader() != TRUE) {
    log("%s: transport not shutdown");
  }
  log("%s: trader number %d terminating", progname, trader_number);
  if (trader_number == 0) {
    /* let command line "parent" close the resource */
    (void)semaphore_close(semaphore);
  }
  exit(0);
}
/*
 * This is the routine that randomly selects from the possible options and
 * performs the work.
 */
static void begin_trading()
{
  log("trader.begin_trading: enter");
  bool_t no_errors = TRUE;
  int counter = 1;
  int choice = 0;
  /* number of choices of messages plus SLEEP  and DISPLAY */
  int possible_choices = (sizeof requests / sizeof(char *) + 2);

  log("possible_choices=%d", possible_choices);

  /* do an initial buy before looping
     -------------------------------- */
  /*  (void)do_buy();*/ /* makes simulation more interesting, but not part of
                           the assignment */

  
  /*    for (count = 1; count < 50; count++) {*/
  while (1) {
    if (no_errors == FALSE) {
      log("trader.begin_trading: no_errors set to FALSE");
      break;
    }

    choice  = random_choice(possible_choices);

    log("**********\t trader=%d iteration=%d Choice=%d\t  **********", 
        trader_number, counter, choice);

    switch (choice) 
      {
      case BUY:
        no_errors = do_buy();
        break;
      case SELL:
        no_errors = do_sell();
        break;
      case CHECK:
        no_errors = do_check();
        break;
      case DISPLAY:
        do_display();
        break;
      case SLEEP:
        log("sleeping...");
        sleep(TRADER_SLEEP_SECONDS);
        break;
      default:
        log("trader.begin_trading: unexpected choice=%d", choice);
        no_errors = TRUE;
        break;
      }
    counter++;
  }
  return;
}

/*
 * Scan our portfolio and return the index into the stock list for the nth
 * stock.
 */
static int find_stock_code(int nth)
{
  log("trader.find_stock_code entry %d", nth);
  int count = 0;
  int symbol_code = -1;
  struct portfolio_entry *current = portfolio;

  while (current != NULL) {
    if (count == nth) break;
    current = current->next;
    count++;
  }

  if (current != NULL) {
    log("trader.find_stock_code: looking for symbol_code for %s",
        current->astock->symbol);
    symbol_code = find_stock_by_name(current->astock->symbol);
  } else {
    log("trader.find_stock_code: could not find entry %d", nth);    
  }
  assert(symbol_code >= 0);
  return symbol_code;
}

/*
 * Dispatch to the appropriate update routine.
 */
static void update_portfolio(int action, int symbol_index, char *msg)
{
  switch (action)
    {
    case BUY:
      log("update_portfolio: BUY %s", symbols[symbol_index]);
      process_buy(symbol_index, msg);
      break;
    case SELL:
      log("update_portfolio: SELL %s", symbols[symbol_index]);
      process_sell(symbol_index, msg);
      break;
    case CHECK:
      log("update_portfolio: CHECK %s", symbols[symbol_index]);
      process_check(symbol_index, msg);
      break;
    default:
      log("%s: unexpected action for trader %d in update_portfolio: <%d>",
          progname, trader_number, action);
      exit(2);
      break;
    }
  return;
}

/*
 * Setup a BUY and send
 */
static bool_t do_buy(void)
{
  log("trader.do_buy: enter");
  bool_t status = TRUE;
  int symbol_index = random_choice(NUMBER_SYMBOLS);
  char msgbuf[MAX_REQUEST_LENGTH + SYMBOL_LENGTH + 2];

  /* Build message to send  and send it
     ---------------------------------- */
  (void)build_request(msgbuf, sizeof msgbuf, symbol_index, BUY);
  log("trader.do_buy: request message = \"%s\"", msgbuf);
  if(send_request(msgbuf) != TRUE)
    status = FALSE;

  if (!(status == FALSE)) {
    memset(result, '\0', sizeof result);
    if (get_result(result, sizeof result) == TRUE) {
      /* Now process the result for our portfolio
         ---------------------------------------- */
      update_portfolio(BUY, symbol_index, result);
    } else {
      status = FALSE;
    }
  }
  log("trader.do_buy: status=%s result=%s",
      (status == TRUE ? "TRUE" : "FALSE"),
      result);
  return status;
}

/*
 * Setup a SELL and send
 */
static bool_t do_sell(void)
{
  log("trader.do_sell: enter");
  bool_t status = TRUE;
  int sell_choice = 0;
  int symbol_index = 0;
  char msgbuf[MAX_REQUEST_LENGTH + SYMBOL_LENGTH + 2];

  /* do we have anything to sell? 
     ---------------------------- */
  if (stock_count > 0) {
    sell_choice = random_choice(stock_count);
    log("sell_choice=%d", sell_choice);

    /* get the index of out sell_choice'th stock 
       ----------------------------------------- */
    symbol_index = find_stock_code(sell_choice);

    /* Build message to send  and send it
       ---------------------------------- */
    (void)build_request(msgbuf, sizeof msgbuf, symbol_index, SELL);
    log("trader.do_sell: request message = \"%s\"", msgbuf);

    if(send_request(msgbuf) != TRUE) {
      status = FALSE;
    } else {
      memset(result, '\0', sizeof result);
      if (get_result(result, sizeof result) == TRUE) {
        log("result: %s", result);
        /* Now process the result for our portfolio
           ---------------------------------------- */
        update_portfolio(SELL, symbol_index, result);
      } else {
        status = FALSE;
      }
    }
    log("trader.do_sell: status=%s result=%s",
        (status == TRUE ? "TRUE" : "FALSE"),
        result);
  } else {
    strcpy(result,"Nothing to sell");
  }
  log("trader.do_sell: status=%s result=%s",
      (status == TRUE ? "TRUE" : "FALSE"),
      result);
  return status;
}

/*
 * Setup a CHECK and send
 */
static bool_t do_check(void)
{
  log("trader.do_check: enter: stock_count=%d", stock_count);
  /*  struct portfolio_entry *mystocks = portfolio;*/
  int check_choice = 0;
  int symbol_index = 0;
  bool_t status = TRUE;
  char msgbuf[MAX_REQUEST_LENGTH + SYMBOL_LENGTH + 2];

  /* If we own stock, check on the price of one
     ------------------------------------------ */
  if (stock_count > 0) {
    check_choice = random_choice(stock_count);
    symbol_index = find_stock_code(check_choice);
    build_request(msgbuf, sizeof msgbuf, symbol_index, CHECK);
    log("trader.do_check: sending \"%s\"", msgbuf);
    if(send_request(msgbuf) != TRUE) {
      status = FALSE;
    } else {
      memset(result, '\0', sizeof result);
      if (get_result(result, sizeof result) == TRUE) {
        log("result: %s", result);
        /* Now process the result for our portfolio
           ---------------------------------------- */
        update_portfolio(CHECK, symbol_index, result);
      } else {
        status = FALSE;
      }
    }
  }
  log("trader.do_check: status=%s",
      (status == TRUE ? "TRUE" : "FALSE"));
  return status;
}
/*
 * Display on standard out the current values of a portfolio.
 */
static void do_display(void)
{
  log("trader.do_display: enter: stock_count=%d", stock_count);
  int count = 1;
  int status = 0;
  double total_value = 0;

  /* Acquire exclusive rights to the screen
     -------------------------------------- */
  status = semaphore_acquire(semaphore);
  assert(status != -1);
  time_t now = time(0);
  printf("************************************************************\n");
  printf("Trader %d Portfolio as of %s", trader_number, ctime(&now));
  struct portfolio_entry *current = portfolio;

  assert(stock_count >= 0);
  if (stock_count > 0) {
    while (count <= stock_count && current != NULL) {
      log("trader.do_display: stock %d, symbol %s", 
          count, current->astock->symbol);
      printf("    Stock: %s\tShares: %5d\tUnit Price: $%6.2f Value: $%8.2f\n",
             current->astock->symbol,
             current->shares,
             (current->astock->unitvalue / 100),
             (current->shares * (current->astock->unitvalue)/100));
      total_value = total_value + 
        (current->shares * ((current->astock->unitvalue)/100));
      current = current->next;
    }
    printf("\tTrader %d portfolio value: $%10.2f\n",
           trader_number, total_value);
  } else {
    log("trader.do_display: No stocks in portfolio!");
    printf("\tNo stocks in Trader %d portfolio!\n", trader_number);
  }
  printf("************************************************************\n");
  fflush(stdout);
  status = semaphore_release(semaphore);
  assert(status != -1);
  log("trader.do_display: return");
  return;
}

/*
 * Process the result of a BUY request
 */
void process_buy(int symbol_index, char *msg)
{
  log("trader.process_buy: enter: symbol_index=%d", symbol_index);
  char *result_part;
  char *price_part;
  int shares_owned = 0;
  double unitvalue = 0;

  result_part = get_result_part(msg);
  log("trader.process_buy: result_part=\"%s\"",result_part);
  if (strcmp(result_part, "YES") == 0) {
    /* process successful BUY */
    price_part = get_check_price_part(msg);
    unitvalue = atof(price_part);
    if (is_owned(portfolio, symbol_index) == YES) {
      /* We already own some obf this */
      shares_owned = update_stock_count(portfolio, symbol_index, 100);
      log("trader.process_buy: shares_owned=%d",shares_owned);
      update_stock_price(portfolio, symbol_index, unitvalue);
    } else {
      /* New to our portfolio */
      portfolio = add_stock(portfolio, symbol_index, 100, unitvalue);
      stock_count = stock_count + 1;
    }
  } else if (strncmp(msg, "DELIST", 6) == 0) {
      /* Oh, NO! Our stock is worthless! */
    log("trader.process_buy: processing DELIST");
    if (is_owned(portfolio, symbol_index) == YES) {
      portfolio = remove_stock(portfolio, symbol_index);
      stock_count = stock_count - 1;
      assert(stock_count >= 0);
    } else {
      log("trader.process_buy: %s not in portfolio", symbols[symbol_index]);
    }
  } else {
  /* The only other option is NO; do nothing for NO */
    log("trader.process_buy: NO matched");
  }
  log("trader.process_buy: return");  
  return;
}

/*
 * Process the result of a SELL request
 */
void process_sell(int symbol_index, char *msg)
{
  log("trader.process_sell: enter: symbol_index=%d", symbol_index);
  char *result_part;
  int remaining_shares = 0;

  result_part = get_result_part(msg);
  log("trader.process_sell: result_part=\"%s\"",result_part);
  if (strcmp(result_part, "YES") == 0) {
    /* Process successful SELL; remove if we have no more of that issue 
       ---------------------------------------------------------------- */
    remaining_shares = update_stock_count(portfolio, symbol_index, -100);
    if (remaining_shares <= 0) {
      log("trader.process_sell: no more shares of %s in portfolio",
          symbols[symbol_index]);
      portfolio = remove_stock(portfolio, symbol_index);
      stock_count = stock_count - 1;
    }
  } else if (strncmp(msg, "DELIST", 6) == 0) {
      /* Oh, NO! Our stock is worthless! */
    log("trader.process_sell: processing DELIST");
    if (is_owned(portfolio, symbol_index) == YES) {
      portfolio = remove_stock(portfolio, symbol_index);
      stock_count = stock_count - 1;
      assert(stock_count >= 0);
    } else {
      log("trader.process_sell: %s not in portfolio", symbols[symbol_index]);
    }
  } else {
  /* The only other option is NO; do nothing for NO */
    log("trader.process_sell: NO MATCH");
  }

  log("trader.process_sell: return");  
  return;
}

/*
 * Process the result of a CHECK request
 */
void process_check(int symbol_index, char *msg)
{
  log("trader.process_check: enter");
  char *result_part;
  char *price_part;
  double unitvalue = 0;

  result_part = get_result_part(msg);
  if (strcmp(result_part, "YES") == 0) {
    log("trader.process_check: successful CHECK");
    /* process successful CHECK */
    price_part = get_check_price_part(msg);
    unitvalue = atof(price_part);
    log("trader.process_check: price_part string \"%s\" as double %.0f",
        price_part, unitvalue);
    update_stock_price(portfolio, symbol_index, unitvalue);
  } else if (strncmp(msg, "DELIST", 6) == 0) {
      /* Oh, NO! Our stock is worthless! */
    log("trader.process_check: DELIST found");
    portfolio = remove_stock(portfolio, symbol_index);
    stock_count = stock_count - 1;
  } else {
  /* The only other option is NO; do nothing for NO */
    log("trader.process_check: NO MATCH");
  }

  log("trader.process_check: return");  
  return;
}

/*
 * Simple help message
 */
static void usage(void)
{
  printf("usage: %s {-h server_address} {-t number_of_traders}\n", progname);
  printf("server_address must be dotted decimal, defaults to 127.0.0.1\n");
  printf("number_of_traders is a decimal number, defaults to 10\n");
}
/* */
