/*
 * title:       Routines that handle message queuing and transport
 *
 * author:      Craig E. Ward
 *
 * envionment:  GCC 3.1
 *
 */
#include <arpa/inet.h>
#include <errno.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "mytypes.h"
#include "transport.h"
#include "util.h"

/* static function prototypes */

/*  external data  */

/* initialized data */
static const unsigned short MARKET_PORT = 51776; /* port for market simulator server */

/* uninitialized data */
static struct sockaddr_in market_server;
static struct sockaddr_in trader_client;
static int sockfd;

/*
 * Create the server socket
 */
bool_t start_server()
{
  bool_t status = TRUE;
  int bindno = 0;
  log("\tstart_server: entered");
  if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) >= 0) {
    /* bind the socket to our server */
    memset(&market_server, '\0', sizeof market_server);

    market_server.sin_family = AF_INET;
    market_server.sin_addr.s_addr = htonl(INADDR_ANY);
    market_server.sin_port = htons(MARKET_PORT);

    bindno = bind(sockfd, (struct sockaddr *)&market_server, sizeof market_server);
    if (bindno < 0) {
      log("\tstart_server: bind error: %s", strerror(errno));
      status = FALSE;
    }

  } else {
    log("\tstart_server: socket error: %s", strerror(errno));
    status = FALSE;
  }
  log("\tstart_server: returning status = %s (sockfd=%d, bindno=%d)",
      (status == TRUE ? "TRUE" : "FALSE"),
      sockfd, bindno);
  return status;
} /* */

/*
 * Read a request from the UDP layer.
 */
request_data_t *get_request(void)
{
  log("\tget_request: enter");
  int addr_len = sizeof(struct sockaddr_in);
  int bytes = 0;
  request_data_t *request_p = (request_data_t *)malloc(sizeof(request_data_t));

  if (request_p != NULL) {
    memset(request_p, '\0', sizeof(request_data_t));
    bytes = recvfrom(sockfd, request_p->request, MAX_MARKET_MSG_BUF,
		     0,
		     (struct sockaddr *)&(request_p->trader_addr),
		     &addr_len);
    log("\tget_request: bytes = %d", bytes);
    if (bytes < 0) {
      log("\tget_request: recvfrom error: %s", strerror(errno));
      free(request_p);
      request_p = NULL;
    } else {
      log("\tget_request: request = \"%s\"", request_p->request);
    }
  } /* if request_p != NULL) */

  log("\tget_request: return %s",(request_p != NULL ? "SUCCESS" : "FAILURE"));
  return request_p;
}

bool_t send_result(request_data_t *request_p)
{
  log("\tsend_result: message = \"%s\"", request_p->result);
  bool_t status = TRUE;

  if (request_p->trader_addr.sin_port == 0) {
    log("\tsend_result: no one to send to!");
    status = FALSE;
  } else  if (sendto(sockfd,
                     request_p->result,
                     (int)strlen(request_p->result),
                     0, 
                     (struct sockaddr *)&(request_p->trader_addr),
                     sizeof(struct sockaddr_in)) == -1) {
    status = FALSE;
    log("\tsend_result: sendto error: %s", strerror(errno));
  }

  free(request_p);

  log("\tsend_result: return %s", (status == TRUE ? "TRUE" : "FALSE"));
  return status;

}

bool_t stop_server()
{
  if (close(sockfd) != -1) return TRUE;
  return FALSE;
}

bool_t stop_trader()
{
  if (close(sockfd) != -1) return TRUE;
  return FALSE;
}
/*
 * Setup the communication link for a trader process to send to the market
 * server.
 */
bool_t init_trader(const char *server_address)
{
  log("\tinit_trader: server address %s",  server_address);
  bool_t status = TRUE;
  int bindno = 0;

  if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) != -1) {
    log("\tinit_trader: sockfd=%d", sockfd);

    memset(&market_server, '\0', sizeof market_server);
    memset(&trader_client, '\0', sizeof trader_client);

    trader_client.sin_family = AF_INET;
    trader_client.sin_addr.s_addr = htonl(INADDR_ANY);
    trader_client.sin_port = htons(0);
    if ((bindno = 
         bind(sockfd,
              (struct sockaddr *)&trader_client,
              sizeof trader_client)) < 0) {
      log("\tinit_trader: bind error: %s", strerror(errno));
      status = FALSE;
    } else {
      market_server.sin_family = AF_INET;
      market_server.sin_port = htons(MARKET_PORT);
      market_server.sin_addr.s_addr = inet_addr(server_address);

      /* Use connect() to inform the UDP software of the destination. If the
       * destination stops receiving, i.e. is shutdown, this will cause a
       * "connection refused" and the trader will exit.
       *
       * Also informs UDP that this is the only source we expect.
       */
      if (connect(sockfd, (struct sockaddr *)&market_server, sizeof market_server) == -1) {
        log("\tinit_trader: connect error: %s", strerror(errno));
        status = FALSE;
      }
    }
  } else {
    log("\tinit_trader: socket failed: %s", strerror(errno));
    status = FALSE;
  }
  log("\tinit_trader: return status = %s",
      (status == TRUE ? "TRUE" : "FALSE"));
  return status;
}

/*
 * Send a request to the market server
 */
bool_t send_request(char *request)
{
  log("\tsend_request: request = \"%s\"", request);
  bool_t status = TRUE;
  int bytes = 0;

  if ((bytes = send(sockfd,
                   request,
                   (int)strlen(request),
                   0)) == -1) {
    status = FALSE;
    log("\tsend_request: send error: %s", strerror(errno));
  }
  log("\tsend_request: return %s", (status == TRUE ? "TRUE" : "FALSE"));
  return status;
}

/*
 * Wait for a reply from the market server.
 */
bool_t get_result(char *result, int buflen) 
{
  log("\tget_result: enter: buflen=%d", buflen);

  fd_set readsocket;
  struct timeval timeout;
  bool_t status = TRUE;
  ssize_t bytes = 0;

  FD_ZERO(&readsocket);
  FD_SET(sockfd, &readsocket);
  timeout.tv_sec = 4;		/* 2 seconds */
  timeout.tv_usec = 0;		/* 0 microseconds */

  switch (select((sockfd+1), &readsocket, (fd_set *)0, (fd_set *)0, &timeout))
    {
    case -1:			/* error */
      log("\tget_result: select error: %s", strerror(errno));
      status = FALSE;
      break;

    case 0:			/* timeout */
      log("\tget_result: select timeout on socket %d", sockfd);
      status = FALSE;
      break;

    default:			/* ready */
      log("\tget_result: select returns read on socket %d", sockfd);
      if ((bytes = recv(sockfd,
			result, 
			buflen,
			0)) < 0) {
	status = FALSE;
	log("\tget_result: recvfrom error: %s", strerror(errno));
      } else {
	log("\tget_result: result = \"%s\", bytes=%.0f", result, bytes);
      }
      break;
    } /* switch */

  log("\tget_result: return %s", (status == TRUE ? "TRUE" : "FALSE"));
  return status;
}
/* */
